@@ -243,11 +243,13 @@ class Class :
             self.methods[method] = None
         else :
             assert methodName is not None, "must supply methodName"
-            self.methods[methodName] = function.Function(method)
+            self.methods[methodName] = function.Function(method, 1)
                 
     def addMethods(self, classObject) :
         for classToken in _getClassTokens(classObject) :
-            token = getattr(classObject, classToken)
+            token = getattr(classObject, classToken, None)
+            if token is None:
+                continue
 
             # Looks like a method.  Need to code it this way to
             # accommodate ExtensionClass and Python 2.2.  Yecchh.
@@ -258,7 +260,7 @@ class Class :
             elif hasattr(token, '__get__') and \
                  not hasattr(token, '__set__') and \
                  type(token) is not types.ClassType :
-                self.addMethod(token.__name__)
+                self.addMethod(getattr(token, '__name__', classToken))
             else :
                 self.members[classToken] = type(token)
                 self.memberRefs[classToken] = None
@@ -271,7 +273,7 @@ class Class :
     def addMembers(self, classObject) :
         if not cfg().onlyCheckInitForMembers :
             for classToken in _getClassTokens(classObject) :
-                method = getattr(classObject, classToken)
+                method = getattr(classObject, classToken, None)
                 if type(method) == types.MethodType :
                     self.addMembersFromMethod(method.im_func)
         else:
@@ -294,7 +296,7 @@ class Class :
                     stack.append(operand)
                 elif OP.STORE_ATTR(op) :
                     if len(stack) > 0 :
-                        if stack[-1] == cfg().methodArgName :
+                        if stack[-1] == cfg().methodArgName:
                             value = None
                             if len(stack) > 1 :
                                 value = type(stack[-2])
@@ -368,12 +370,19 @@ def importError(moduleName):
         e = exc_value
         msg = '%s (%s, line %d)' % (e.msg, e.filename, e.lineno)
         line = _getLineInFile(moduleName, e.lineno)
-        exc_value = '%s\n    %s\n   %s^' % (msg, line, ' ' * e.offset)
+        offset = e.offset
+        if type(offset) is not types.IntType:
+            offset = 0
+        exc_value = '%s\n    %s\n   %s^' % (msg, line, ' ' * offset)
     elif exc_type is not ImportError:
         sys.stderr.write("  Caught exception importing module %s:\n" %
                          moduleName)
 
-        tbinfo = traceback.extract_tb(tb)
+        try:
+            tbinfo = traceback.extract_tb(tb)
+        except:
+            tbinfo = []
+            sys.stderr.write("      Unable to format traceback\n")
         for filename, line, func, text in tbinfo[1:]:
             sys.stderr.write("    File \"%s\", line %d" % (filename, line))
             if func != "?":
@@ -391,6 +400,14 @@ def importError(moduleName):
         sys.stderr.write('**error formatting exception value**\n')
 
 
+def _getPyFile(filename):
+    """Return the file and '.py' filename from a filename which could
+    end with .py, .pyc, or .pyo"""
+
+    if filename[-1] in 'oc' and filename[-4:-1] == '.py':
+        return filename[:-1]
+    return filename
+
 class Module :
     "Class to hold all information for a module"
 
@@ -426,8 +443,14 @@ class Module :
 
     def addClass(self, name) :
         self.classes[name] = c = Class(name, self.module)
-        packages = string.split(str(c.classObject), '.')
-        c.ignoreAttrs = packages[0] in cfg().blacklist
+        try:
+            objName = str(c.classObject)
+        except TypeError:
+            # this can happen if there is a goofy __getattr__
+            c.ignoreAttrs = 1
+        else:
+            packages = string.split(objName, '.')
+            c.ignoreAttrs = packages[0] in cfg().blacklist
         if not c.ignoreAttrs :
             self.__addAttributes(c, c.classObject)
 
@@ -445,26 +468,22 @@ class Module :
             self.modules[name] = module
 
     def filename(self) :
-        if not self.module :
-            return self.moduleName
         try :
             filename = self.module.__file__
         except AttributeError :
             filename = self.moduleName
-        if string.lower(filename[-4:]) == '.pyc' :
-            filename = filename[:-4] + '.py'
-        return filename
+        return _getPyFile(filename)
 
-    def load(self) :
+    def load(self):
         try :
             # there's no need to reload modules we already have
             module = sys.modules.get(self.moduleName)
             if module :
                 if not _allModules[self.moduleName].module :
-                    return self.initModule(module)
+                    return self._initModule(module)
                 return 1
 
-            return self.initModule(self.setupMainCode())
+            return self._initModule(self.setupMainCode())
         except (SystemExit, KeyboardInterrupt) :
             exc_type, exc_value, exc_tb = sys.exc_info()
             raise exc_type, exc_value
@@ -473,6 +492,19 @@ class Module :
             return 0
 
     def initModule(self, module) :
+        if not self.module:
+            filename = _getPyFile(module.__file__)
+            if string.lower(filename[-3:]) == '.py':
+                try:
+                    file = open(filename)
+                except IOError:
+                    pass
+                else:
+                    self._setupMainCode(file, filename, module)
+            return self._initModule(module)
+        return 1
+
+    def _initModule(self, module):
         self.module = module
         self.attributes = dir(self.module)
 
@@ -501,15 +533,17 @@ class Module :
     def setupMainCode(self) :
         file, filename, smt = _findModule(self.moduleName)
         # FIXME: if the smt[-1] == imp.PKG_DIRECTORY : load __all__
+        module = imp.load_module(self.moduleName, file, filename, smt)
+        self._setupMainCode(file, filename, module)
+        return module
+
+    def _setupMainCode(self, file, filename, module):
         try :
-            module = imp.load_module(self.moduleName, file, filename, smt)
             self.main_code = function.create_from_file(file, filename, module)
         finally :
             if file != None :
                 file.close()
 
-        return module
-
 
 def getAllModules() :
     "Returns a list of all modules that should be checked."
@@ -525,8 +559,10 @@ _BUILTIN_MODULE_ATTRS = { 'sys': [ 'ps1', 'ps2', 'tracebacklimit',
                                  ],
                         }
 
-def fixupBuiltinModules() :
+def fixupBuiltinModules(needs_init=0):
     for moduleName in sys.builtin_module_names :
+        if needs_init:
+            _ = Module(moduleName, 0)
         module = _allModules.get(moduleName, None)
         if module is not None :
             try :
@@ -597,6 +633,26 @@ def main(argv) :
         sys.stderr.write(_VERSION_MISMATCH_ERROR)
         sys.exit(100)
 
+    # remove empty arguments
+    argv = filter(None, argv)
+        
+    # if the first arg starts with an @, read options from the file
+    # after the @ (this is mostly for windows)
+    if len(argv) >= 2 and argv[1][0] == '@':
+        # read data from the file
+        command_file = argv[1][1:]
+        try:
+            f = open(command_file, 'r')
+            command_line = f.read()
+            f.close()
+        except IOError, err:
+            sys.stderr.write("Unable to read commands from file: %s\n  %s\n" % \
+                             (command_file, err))
+            sys.exit(101)
+
+        # convert to an argv list, keeping argv[0] and the files to process
+        argv = argv[:1] + string.split(command_line) + argv[2:]
+ 
     global _cfg
     _cfg, files, suppressions = Config.setupFromArgs(argv[1:])
     if not files :
@@ -632,15 +688,33 @@ if __name__ == '__main__' :
 else :
     _orig__import__ = None
     _suppressions = None
+    _warnings_cache = {}
+
+    def _get_unique_warnings(warnings):
+        for i in range(len(warnings)-1, -1, -1):
+            w = warnings[i].format()
+            if _warnings_cache.has_key(w):
+                del warnings[i]
+            else:
+                _warnings_cache[w] = 1
+        return warnings
+
+    def __import__(name, globals=None, locals=None, fromlist=None):
+        if globals is None:
+            globals = {}
+        if locals is None:
+            locals = {}
+        if fromlist is None:
+            fromlist = []
 
-    def __import__(name, globals={}, locals={}, fromlist=[]) :
         check = not sys.modules.has_key(name) and name[:10] != 'pychecker.'
         pymodule = _orig__import__(name, globals, locals, fromlist)
         if check :
             try :
                 module = Module(pymodule.__name__)
-                if module.load() :
-                    _printWarnings(warn.find([module], _cfg, _suppressions))
+                if module.initModule(pymodule):
+                    warnings = warn.find([module], _cfg, _suppressions)
+                    _printWarnings(_get_unique_warnings(warnings))
                 else :
                     print 'Unable to load module', pymodule.__name__
             except Exception:
@@ -655,7 +729,7 @@ else :
         args = string.split(os.environ.get('PYCHECKER', ''))
         _cfg, files, _suppressions = Config.setupFromArgs(args)
         utils.initConfig(_cfg)
-        fixupBuiltinModules()
+        fixupBuiltinModules(1)
 
         # keep the orig __import__ around so we can call it
         import __builtin__
