@@ -15,14 +15,17 @@ import time
 
 _RC_FILE = ".pycheckrc"
 CHECKER_VAR = '__pychecker__'
-_VERSION = '0.8.11'
+_VERSION = '0.8.13'
 
 _DEFAULT_BLACK_LIST = [ "Tkinter", "wxPython", "gtk", "GTK", "GDK", ]
 _DEFAULT_VARIABLE_IGNORE_LIST = [ '__version__', '__warningregistry__', 
-                                  '__all__', '__credits__',
-                                  '__author__', '__email__', ]
+                                  '__all__', '__credits__', '__test__',
+                                  '__author__', '__email__', '__revision__', ]
 _DEFAULT_UNUSED_LIST = [ '_', 'empty', 'unused', 'dummy', ]
 
+# All these options are on even if -e/--errors is used
+_ERRORS = { 'noEffect': 1, }
+
 _OPTIONS = (
     ('Major Options', [
  ('e', 0, 'errors', None, 'turn off all warnings which are not likely errors'),
@@ -54,10 +57,12 @@ _OPTIONS = (
  ('A', 0, 'callattr', 'callingAttribute', 'Calling data members as functions'),
  ('y', 0, 'classattr', 'classAttrExists', 'class attribute does not exist'),
  ('S', 1, 'self', 'methodArgName', 'First argument to methods'),
+ ('',  1, 'classmethodargs', 'classmethodArgNames', 'First argument to classmethods'),
  ('T', 0, 'argsused', 'argumentsUsed', 'unused method/function arguments'),
  ('z', 0, 'varargsused', 'varArgumentsUsed', 'unused method/function variable arguments'),
  ('G', 0, 'selfused', 'ignoreSelfUnused', 'ignore if self is unused in methods'),
  ('o', 0, 'override', 'checkOverridenMethods', 'check if overridden methods have the same signature'),
+ ('',  0, 'special', 'checkSpecialMethods', 'check if __special__ methods exist and have the correct signature'),
  ('U', 0, 'reuseattr', 'redefiningFunction', 'check if function/class/method names are reused'),
  ('Y', 0, 'positive', 'unaryPositive', 'check if using unary positive (+) which is usually meaningless'),
  ('j', 0, 'moddefvalue', 'modifyDefaultValue', 'check if modify (call method) on a parameter that has a default value'),
@@ -65,6 +70,9 @@ _OPTIONS = (
  ( '', 0, 'unpack', 'unpackNonSequence', 'check if unpacking a non-sequence'),
  ( '', 0, 'unpacklen', 'unpackLength', 'check if unpacking sequence with the wrong length'),
  ( '', 0, 'badexcept', 'badExceptions', 'check if raising or catching bad exceptions'),
+ ('4', 0, 'noeffect', 'noEffect', 'check if statement appears to have no effect'),
+ ('',  0, 'modulo1', 'modulo1', 'check if using (expr % 1), it has no effect on integers and strings'),
+ ('',  0, 'isliteral', 'isLiteral', "check if using (expr is const-literal), doesn't always work on integers and strings"),
      ]),
     ('Possible Errors', [
  ('r', 0, 'returnvalues', 'checkReturnValues', 'check consistent return values'),
@@ -75,6 +83,7 @@ _OPTIONS = (
  ( '', 0, 'emptyslots', 'emptySlots', 'check if __slots__ is empty'),
  ('D', 0, 'intdivide', 'intDivide', 'check if using integer division'),
  ('w', 0, 'shadow', 'shadows', 'check if local variable shadows a global'),
+ ('s', 0, 'shadowbuiltin', 'shadowBuiltins', 'check if a variable shadows a builtin'),
      ]),
     ('Security', [
  ( '', 0, 'input', 'usesInput', 'check if input() is used'),
@@ -223,14 +232,19 @@ class Config :
         self.unpackNonSequence = 1
         self.unpackLength = 1
         self.badExceptions = 1
+        self.noEffect = 1
         self.deprecated = 1
+        self.modulo1 = 1
+        self.isLiteral = 1
 
         self.unusedNames = _DEFAULT_UNUSED_LIST
         self.variablesToIgnore = _DEFAULT_VARIABLE_IGNORE_LIST
         self.blacklist = _DEFAULT_BLACK_LIST
         self.ignoreStandardLibrary = 0
         self.methodArgName = 'self'
+        self.classmethodArgNames = ['cls', 'klass']
         self.checkOverridenMethods = 1
+        self.checkSpecialMethods = 1
 
         self.argumentsUsed = 1
         self.varArgumentsUsed = 1
@@ -252,10 +266,12 @@ class Config :
         self.checkImplicitReturns = 1
         self.intDivide = 1
         self.shadows = 1
+        self.shadowBuiltins = 1
         self.unaryPositive = 1
         self.modifyDefaultValue = 1
         self.usesExec = 0
         self.usesInput = 1
+        self.constAttr = 1
 
     def loadFile(self, filename) :
         suppressions = {}
@@ -266,7 +282,8 @@ class Config :
             for key, value in dict.items() :
                 if self.__dict__.has_key(key) :
                     self.__dict__[key] = value
-                elif key not in ('suppressions', 'suppressionRegexs') :
+                elif key not in ('suppressions', 'suppressionRegexs') and \
+                     key[0] != '_':
                     print "Warning, option (%s) doesn't exist, ignoring" % key
 
             suppressions = _getSuppressions('suppressions', dict, filename)
@@ -332,6 +349,14 @@ class Config :
                     newValue = int(newValue)
                 elif memberType == type([]) :
                     newValue = string.split(newValue, ',')
+                elif memberType == type('') and \
+                     newValue[0] in '\'"':
+                        try:
+                            newValue = eval(newValue)
+                        except:
+                            msg = 'Invalid option parameter: %s for %s\n' % \
+                                  (`newValue`, arg)
+                            sys.stderr.write(msg)
                 setattr(self, member, newValue)
             elif arg[0:2] == '--' :
                 setattr(self, member, arg[2:5] != 'no-')
@@ -349,7 +374,7 @@ def errors_only(complexity = 0) :
     "Return {} of Config with all warnings turned off"
     dict = Config().__dict__
     for k, v in dict.items() :
-        if type(v) == type(0) and v >= complexity :
+        if type(v) == type(0) and v >= complexity and not _ERRORS.has_key(k):
             dict[k] = 0
     return dict
 
