@@ -22,6 +22,18 @@ __pychecker__ = 'no-argsused'
 def cfg() :
     return utils.cfg()
 
+def getFunctionArgErr(func_name, argCount, minArgs, maxArgs):
+    err = None
+    if maxArgs == None:
+        if argCount < minArgs :
+            err = msgs.INVALID_ARG_COUNT2 % (func_name, argCount, minArgs)
+    elif argCount < minArgs or argCount > maxArgs:
+        if minArgs == maxArgs:
+            err = msgs.INVALID_ARG_COUNT1 % (func_name, argCount, minArgs)
+        else:
+            err = msgs.INVALID_ARG_COUNT3 % (func_name, argCount, minArgs, maxArgs)
+    return err
+
 def _checkFunctionArgCount(code, func_name, argCount, minArgs, maxArgs,
                            objectReference = 0) :
     # there is an implied argument for object creation and self.xxx()
@@ -30,16 +42,7 @@ def _checkFunctionArgCount(code, func_name, argCount, minArgs, maxArgs,
         if maxArgs is not None :
             maxArgs = maxArgs - 1
 
-    err = None
-    if maxArgs == None :
-        if argCount < minArgs :
-            err = msgs.INVALID_ARG_COUNT2 % (func_name, argCount, minArgs)
-    elif argCount < minArgs or argCount > maxArgs :
-        if minArgs == maxArgs :
-            err = msgs.INVALID_ARG_COUNT1 % (func_name, argCount, minArgs)
-        else :
-            err = msgs.INVALID_ARG_COUNT3 % (func_name, argCount, minArgs, maxArgs)
-
+    err = getFunctionArgErr(func_name, argCount, minArgs, maxArgs)
     if err :
         code.addWarning(err)
 
@@ -47,16 +50,14 @@ def _checkFunctionArgs(code, func, objectReference, argCount, kwArgs,
                        check_arg_count = 1) :
     func_name = func.function.func_code.co_name
     if kwArgs :
-        func_args = func.function.func_code.co_varnames
-        func_args_len = len(func_args)
-        if argCount < func_args_len and kwArgs[0] in func_args[argCount:] :
+        args_len = func.function.func_code.co_argcount
+        arg_names = func.function.func_code.co_varnames[argCount:args_len]
+        if argCount < args_len and kwArgs[0] in arg_names:
             if cfg().namedArgs :
                 code.addWarning(msgs.FUNC_USES_NAMED_ARGS % func_name)
 
             # convert the named args into regular params, and really check
-            origArgCount = argCount
-            while kwArgs and argCount < func_args_len and \
-                  kwArgs[0] in func_args[origArgCount:] :
+            while argCount < args_len and kwArgs and kwArgs[0] in arg_names:
                 argCount = argCount + 1
                 kwArgs = kwArgs[1:]
             _checkFunctionArgs(code, func, objectReference, argCount, kwArgs,
@@ -111,7 +112,7 @@ def _getFunction(module, stackValue) :
         return _getReferenceFromModule(module, identifier[-1])
 
     # we can't handle self.x.y
-    if (i+2) == maxLen and identifier[0] == cfg().methodArgName :
+    if (i+2) == maxLen and identifier[0] == cfg().methodArgName:
         return None, None, 0
 
     c = module.classes.get(identifier[-2], None)
@@ -127,8 +128,20 @@ def _checkBuiltin(code, loadValue, argCount, kwArgs, check_arg_count = 1) :
         if info is not None :
             if func_name == 'input' and cfg().usesInput:
                 code.addWarning(msgs.USES_INPUT)
-            if kwArgs :
-                code.addWarning(msgs.FUNC_DOESNT_SUPPORT_KW % func_name)
+            if cfg().constAttr and \
+               ((func_name == 'setattr' and argCount >= 2) or 
+                (func_name == 'getattr' and argCount == 2)):
+                arg2 = code.stack[-argCount + 1]
+                if arg2.const:
+                    code.addWarning(msgs.USES_CONST_ATTR % func_name)
+
+            if kwArgs:
+                if len(info) < 4:
+                    code.addWarning(msgs.FUNC_DOESNT_SUPPORT_KW % func_name)
+                elif info[3]:
+                    for arg in kwArgs:
+                        if arg not in info[3]:
+                            code.addWarning(msgs.FUNC_DOESNT_SUPPORT_KW_ARG % (func_name, arg))
             elif check_arg_count :
                 _checkFunctionArgCount(code, func_name, argCount,
                                        info[1], info[2])
@@ -184,8 +197,13 @@ def _checkModifyDefaultArg(code, objectName, methodName=None) :
 def _isexception(object) :
     # FIXME: i have no idea why this function is necessary
     # it seems that the issubclass() should work, but it doesn't always
-    if issubclass(object, Exception) :
-        return 1
+    try:
+        # try/except is necessary for globals like NotImplemented
+        if issubclass(object, Exception) :
+            return 1
+    except TypeError:
+        return 0
+
     for c in object.__bases__ :
         if utils.startswith(str(c), 'exceptions.') :
             return 1
@@ -265,7 +283,7 @@ def _handleFunctionCall(codeSource, code, argCount, indexOffset = 0,
     funcName = loadValue.getName()
     returnValue = Stack.makeFuncReturnValue(loadValue, argCount)
 
-    if loadValue.isMethodCall(codeSource.classObject, cfg().methodArgName) :
+    if loadValue.isMethodCall(codeSource.classObject, cfg().methodArgName):
         methodName = loadValue.data[1]
         try :
             m = codeSource.classObject.methods[methodName]
@@ -310,6 +328,8 @@ def _handleFunctionCall(codeSource, code, argCount, indexOffset = 0,
                 _checkAbstract(refClass, code, funcName)
 
             if func != None :
+                if refClass and func.isClassMethod():
+                    argCount = argCount + 1
                 _checkFunctionArgs(code, func, method, argCount, kwArgs,
                                    check_arg_count)
                 # if this isn't a c'tor, we should check
@@ -320,11 +340,15 @@ def _handleFunctionCall(codeSource, code, argCount, indexOffset = 0,
                     if method :
                         # c'tor, return the class as the type
                         returnValue = Stack.Item(loadValue, refClass)
-                
-                    elif argCount > 0 and \
+                    elif func.isClassMethod():
+                        # FIXME: do anything here?
+                        pass
+                    elif argCount > 0 and cfg().methodArgName and \
+                         not func.isStaticMethod() and \
                          code.stack[funcIndex].type == Stack.TYPE_ATTRIBUTE and \
-                         code.stack[funcIndex+1].data != cfg().methodArgName :
-                        code.addWarning(msgs.SELF_NOT_FIRST_ARG % cfg().methodArgName)
+                         code.stack[funcIndex+1].data != cfg().methodArgName:
+                        e = msgs.SELF_NOT_FIRST_ARG % (cfg().methodArgName, '')
+                        code.addWarning(e)
             elif refClass and method :
                 returnValue = Stack.Item(loadValue, refClass)
                 if (argCount > 0 or len(kwArgs) > 0) and \
@@ -391,12 +415,15 @@ def _getGlobalName(name, func) :
     return name
 
 
+def _checkNoEffect(code):
+    if OP.POP_TOP(code.nextOpInfo()[0]) and cfg().noEffect:
+        code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
+    
 def _makeConstant(code, index, factoryFunction) :
     "Build a constant on the stack ((), [], or {})"
     if index > 0 :
         code.stack[-index:] = [ factoryFunction(code.stack[-index:]) ]
-        if OP.POP_TOP(code.nextOpInfo()[0]) :
-            code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
+        _checkNoEffect(code)
     else :
         code.pushStack(factoryFunction())
 
@@ -463,7 +490,7 @@ def _handleImport(code, operand, module, main, fromName) :
                 err = 'from %s import %s' % (tmpFromName, operand)
                 err = msgs.MODULE_MEMBER_IMPORTED_AGAIN % err
             elif modline1 is not None :
-                if cfg().mixImport and code.lastLineNum != modline1[1] :
+                if cfg().mixImport and code.getLineNum() != modline1[1] :
                     err = msgs.MIX_IMPORT_AND_FROM_IMPORT % tmpFromName
             else :
                 err = msgs.MODULE_MEMBER_ALSO_STAR_IMPORTED % fromName
@@ -476,7 +503,7 @@ def _handleImport(code, operand, module, main, fromName) :
                 code.addWarning(err)
 
     if main :
-        fileline = (code.func_code.co_filename, code.lastLineNum)
+        fileline = (code.func_code.co_filename, code.getLineNum())
         module.moduleLineNums[key] = fileline
         if fromName is not None :
             module.moduleLineNums[(fromName,)] = fileline
@@ -508,6 +535,11 @@ def _getFormatInfo(format, code) :
     # skip the first item in the list, it's always empty
     for section in sections[1:] :
         orig_section = section
+        if not section:
+            code.addWarning(msgs.INVALID_FORMAT % orig_section +
+                            ' (end of format string)')
+            continue
+
         # handle dictionary formats
         if section[0] == '(' :
             mappingFormatCount = mappingFormatCount + 1
@@ -545,8 +577,16 @@ def _getFormatInfo(format, code) :
 
     return formatCount, vars
 
+try:
+    unicode, UnicodeError
+except NameError:
+    UnicodeError = None
+
 def _getConstant(code, module, data) :
-    data = str(data.data)
+    try:
+        data = str(data.data)
+    except UnicodeError:
+        data = unicode(data.data)
     format = code.constants.get(data)
     if format is not None :
         return format
@@ -649,6 +689,18 @@ def _getTypeStr(t):
     return returnStr
 
 
+def _getLineNum(co, instr_index):
+    co_lnotab = co.co_lnotab
+    lineno = co.co_firstlineno
+    addr = 0
+    for lnotab_index in range(0, len(co_lnotab), 2):
+        addr = addr + ord(co_lnotab[lnotab_index])
+        if addr > instr_index:
+            return lineno
+        lineno = lineno + ord(co_lnotab[lnotab_index+1])
+    return lineno
+
+
 class Code :
     'Hold all the code state information necessary to find warnings'
 
@@ -694,9 +746,17 @@ class Code :
             self.unusedLocals[arg] = 0
             self.typeMap[arg] = [ Stack.TYPE_UNKNOWN ]
 
+    def getLineNum(self):
+        line = self.lastLineNum
+        # if we don't have linenum info, calc it from co_lntab & index
+        if line == self.func_code.co_firstlineno:
+            # FIXME: this could be optimized, if we kept last line info
+            line = _getLineNum(self.func_code, self.index - 1)
+        return line
+
     def getWarning(self, err, line = None) :
         if line is None :
-            line = self.lastLineNum
+            line = self.getLineNum()
         return Warning.Warning(self.func_code, line, err)
 
     def addWarning(self, err, line = None) :
@@ -739,8 +799,7 @@ class Code :
 
     def pushStack(self, item) :
         self.stack.append(item)
-        if OP.POP_TOP(self.nextOpInfo()[0]) :
-            self.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
+        _checkNoEffect(self)
 
     def popStack(self) :
         if self.stack :
@@ -809,12 +868,12 @@ class Code :
 
     def addReturn(self) :
         if len(self.stack) > 0 :
-            value = (self.lastLineNum, self.stack[-1], self.nextOpInfo()[2])
+            value = (self.getLineNum(), self.stack[-1], self.nextOpInfo()[2])
             self.returnValues.append(value)
             self.popStack()
 
     def addRaise(self) :
-        self.raiseValues.append((self.lastLineNum, None, self.nextOpInfo()[2]))
+        self.raiseValues.append((self.getLineNum(), None, self.nextOpInfo()[2]))
 
     def addBranch(self, label) :
         if label is not None :
@@ -843,11 +902,11 @@ class Code :
         rc = utils.shouldUpdateArgs(operand)
         if rc :
             utils.updateCheckerArgs(self.stack[-1].data, self.func_code,
-                                    self.lastLineNum, self.warnings)
+                                    self.getLineNum(), self.warnings)
         return rc
         
     def updateModuleLineNums(self, module, operand) :
-        filelist = (self.func_code.co_filename, self.lastLineNum)
+        filelist = (self.func_code.co_filename, self.getLineNum())
         module.moduleLineNums[operand] = filelist
 
 
@@ -875,6 +934,10 @@ def _checkAssign(code, name):
         if cap in _BAD_ASSIGN_NAMES:
             code.addWarning(msgs.SHOULDNT_ASSIGN_NAME % (name, cap))
 
+def _checkVariableOperationOnItself(code, lname, msg):
+    if code.stack and code.stack[-1].getName() == lname:
+        code.addWarning(msg % lname)
+
 def _checkFutureKeywords(code, varname) :
     kw = python.FUTURE_KEYWORDS.get(varname)
     if kw is not None :
@@ -885,13 +948,14 @@ def _STORE_NAME(oparg, operand, codeSource, code) :
         _checkFutureKeywords(code, operand)
         module = codeSource.module
         if not codeSource.in_class :
+            _checkShadowBuiltin(code, operand)
             if not codeSource.calling_code :
                 _checkGlobal(operand, module, codeSource.func, code,
                              msgs.GLOBAL_DEFINED_NOT_DECLARED, codeSource.main)
         else :
             if code.stack :
                 codeSource.classObject.statics[operand] = code.stack[-1]
-                codeSource.classObject.lineNums[operand] = code.lastLineNum
+                codeSource.classObject.lineNums[operand] = code.getLineNum()
 
         var = module.variables.get(operand)
         if var is not None and code.stack and code.stack[-1].const :
@@ -938,7 +1002,7 @@ def _LOAD_NAME(oparg, operand, codeSource, code) :
     # if there was from XXX import *, _* names aren't imported
     if codeSource.module.modules.has_key(operand) and \
        hasattr(codeSource.module.module, operand) :
-        operand = eval("codeSource.module.module.%s.__name__" % operand)
+        operand = getattr(codeSource.module.module, operand).__name__
 
     opType, const = Stack.TYPE_GLOBAL, 0
     if operand == 'None' :
@@ -986,13 +1050,17 @@ def _LOAD_CONST(oparg, operand, codeSource, code) :
 
 
 def _checkLocalShadow(code, module, varname) :
-    if module.variables.has_key(varname) :
+    if module.variables.has_key(varname) and cfg().shadows :
         line = module.moduleLineNums.get(varname, ('<unknown>', 0))
-        if cfg().shadows :
-            w = code.getWarning(msgs.LOCAL_SHADOWS_GLOBAL % (varname, line[1]))
-            if line[0] != w.file:
-                w.err = '%s in file %s' % (w.err, line[0])
-            code.addWarning(w)
+        w = code.getWarning(msgs.LOCAL_SHADOWS_GLOBAL % (varname, line[1]))
+        if line[0] != w.file:
+            w.err = '%s in file %s' % (w.err, line[0])
+        code.addWarning(w)
+
+def _checkShadowBuiltin(code, varname) :
+    if __builtins__.has_key(varname) and varname[0] != '_' and \
+       cfg().shadowBuiltins:
+        code.addWarning(msgs.VARIABLE_SHADOWS_BUILTIN % varname)
 
 def _checkLoadLocal(code, codeSource, varname, deletedWarn, usedBeforeSetWarn) :
     _checkFutureKeywords(code, varname)
@@ -1016,6 +1084,9 @@ def _LOAD_FAST(oparg, operand, codeSource, code) :
 def _STORE_FAST(oparg, operand, codeSource, code) :
     if not code.updateCheckerArgs(operand) :
         _checkFutureKeywords(code, operand)
+        if code.stack and code.stack[-1].type == types.StringType:
+            _checkVariableOperationOnItself(code, operand,
+                                            msgs.SET_VAR_TO_ITSELF)
         code.setType(operand)
         if not code.unpackCount and code.stack and \
            (code.stack[-1].const or code.stack[-1].type == types.TupleType) :
@@ -1025,12 +1096,13 @@ def _STORE_FAST(oparg, operand, codeSource, code) :
                 code.constants[operand] = code.stack[-1].data
 
         _checkLocalShadow(code, codeSource.module, operand)
+        _checkShadowBuiltin(code, operand)
         _checkAssign(code, operand)
         _checkException(code, operand)
         if code.deletedLocals.has_key(operand) :
             del code.deletedLocals[operand]
         if not code.unusedLocals.has_key(operand) :
-            errLine = code.lastLineNum
+            errLine = code.getLineNum()
             if code.unpackCount and not cfg().unusedLocalTuple :
                 errLine = -errLine
             code.unusedLocals[operand] = errLine
@@ -1039,7 +1111,7 @@ def _STORE_FAST(oparg, operand, codeSource, code) :
 def _DELETE_FAST(oparg, operand, codeSource, code) :
     _checkLoadLocal(code, codeSource, operand,
                     msgs.LOCAL_ALREADY_DELETED, msgs.VAR_DELETED_BEFORE_SET)
-    code.deletedLocals[operand] = code.lastLineNum
+    code.deletedLocals[operand] = code.getLineNum()
 
 def _checkAttribute(top, operand, codeSource, code) :
     if top.data == cfg().methodArgName and codeSource.classObject != None :
@@ -1059,7 +1131,7 @@ def _checkExcessiveReferences(code, top, extraAttr = None) :
             data = data + (extraAttr,)
         
         maxReferences = cfg().maxReferences
-        if data[0] == cfg().methodArgName :
+        if data[0] == cfg().methodArgName:
             maxReferences = maxReferences + 1
         if len(data) > maxReferences :
             name = string.join(top.data, '.')
@@ -1101,7 +1173,7 @@ def _LOAD_ATTR(oparg, operand, codeSource, code) :
 
         nextOp = code.nextOpInfo()[0]
         if not OP.LOAD_ATTR(nextOp) :
-            if OP.POP_TOP(nextOp) :
+            if OP.POP_TOP(nextOp) and cfg().noEffect:
                 code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
             else :
                 _checkExcessiveReferences(code, top)
@@ -1113,7 +1185,12 @@ def _ok_to_set_attr(classObject, basename, attr) :
 
 def _STORE_ATTR(oparg, operand, codeSource, code) :
     if code.stack :
-        top = code.stack[-1]
+        top = code.stack.pop()
+        top_name = '%s.%s' % (top.getName(), operand)
+        if top.type in (types.StringType, Stack.TYPE_ATTRIBUTE) and \
+           code.stack[-1].type == Stack.TYPE_ATTRIBUTE:
+            _checkVariableOperationOnItself(code, top_name,
+                                            msgs.SET_VAR_TO_ITSELF)
         _checkExcessiveReferences(code, top, operand)
         if _ok_to_set_attr(codeSource.classObject, top.data, operand) :
             code.addWarning(msgs.INVALID_SET_CLASS_ATTR % operand)
@@ -1136,7 +1213,11 @@ def _getExceptionInfo(codeSource, item):
             pass
 
     if not e:
-        c = codeSource.module.classes.get(item.data)
+        try:
+            c = codeSource.module.classes.get(item.data)
+        except TypeError:     # item.data may not be hashable (e.g., list)
+            return e, 0
+
         if c is not None:
             e = c.classObject
         else:
@@ -1178,8 +1259,9 @@ def _checkBoolean(code, checks):
             data = string.capitalize(item.data)
             if item.type is Stack.TYPE_GLOBAL and data in _BOOL_NAMES:
                 code.addWarning(msgs.BOOL_COMPARE % item.data)
-        except AttributeError:
-            pass # ignore items that are not a StackItem
+        except (AttributeError, TypeError):
+            # TypeError is necessary for Python 1.5.2
+            pass # ignore items that are not a StackItem or a string
 
 def _COMPARE_OP(oparg, operand, codeSource, code) :
     compareValues = _handleComparison(code.stack, operand)
@@ -1187,8 +1269,19 @@ def _COMPARE_OP(oparg, operand, codeSource, code) :
         _handleExceptionChecks(codeSource, code, compareValues)
     elif oparg < OP.IS_COMPARISON:
         _checkBoolean(code, compareValues)
-    if OP.POP_TOP(code.nextOpInfo()[0]) :
-        code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
+    elif cfg().isLiteral:
+        # X is Y   or   X is not Y   comparison
+        second_arg = code.stack[-1].data[2]
+        if second_arg.const:
+            data = second_arg.data
+            if second_arg.type is types.DictType:
+                data = {}
+            not_str = ''
+            if oparg != OP.IS_COMPARISON:
+                not_str = ' not'
+            code.addWarning(msgs.IS_LITERAL % (not_str, data))
+
+    _checkNoEffect(code)
 
 def _IMPORT_NAME(oparg, operand, codeSource, code) :
     code.pushStack(Stack.Item(operand, types.ModuleType))
@@ -1243,7 +1336,9 @@ def _MAKE_FUNCTION(oparg, operand, codeSource, code) :
     newValue = Stack.makeFuncReturnValue(code.stack[-1], oparg)
     code.popStackItems(oparg+1)
     code.pushStack(newValue)
-_MAKE_CLOSURE = _MAKE_FUNCTION
+
+def _MAKE_CLOSURE(oparg, operand, codeSource, code) :
+    _MAKE_FUNCTION(max(0, oparg - 1), operand, codeSource, code)
 
 def _BUILD_MAP(oparg, operand, codeSource, code) :
     _makeConstant(code, oparg, Stack.makeDict)
@@ -1256,9 +1351,19 @@ def _BUILD_CLASS(oparg, operand, codeSource, code) :
     newValue = Stack.makeFuncReturnValue(code.stack[-1], types.ClassType)
     code.popStackItems(3)
     code.pushStack(newValue)
-    
+
+def _modifyStackName(code, suffix):
+    if code.stack:
+        tos = code.stack[-1]
+        tos_type = type(tos.data)
+        if tos_type == types.StringType:
+            tos.data = tos.data + suffix
+        elif tos_type == types.TupleType and \
+             type(tos.data[-1]) == types.StringType:
+            tos.data = tos.data[:-1] + (tos.data[-1] + suffix,)
+
 def _UNARY_CONVERT(oparg, operand, codeSource, code) :
-    if len(code.stack) > 0 :
+    if code.stack:
         stackValue = code.stack[-1]
         if stackValue.data == cfg().methodArgName and \
            stackValue.const == 0 and codeSource.classObject is not None and \
@@ -1266,6 +1371,7 @@ def _UNARY_CONVERT(oparg, operand, codeSource, code) :
             code.addWarning(msgs.USING_SELF_IN_REPR)
         stackValue.data = str(stackValue.data)
         stackValue.type = types.StringType
+    _modifyStackName(code, '-repr')
 
 def _UNARY_POSITIVE(oparg, operand, codeSource, code) :
     if OP.UNARY_POSITIVE(code.nextOpInfo()[0]) :
@@ -1273,28 +1379,53 @@ def _UNARY_POSITIVE(oparg, operand, codeSource, code) :
         code.popNextOp()
     elif cfg().unaryPositive and code.stack and not code.stack[-1].const :
         code.addWarning(msgs.UNARY_POSITIVE_HAS_NO_EFFECT)
+    _modifyStackName(code, '-pos')
 
 def _UNARY_NEGATIVE(oparg, operand, codeSource, code) :
     if OP.UNARY_NEGATIVE(code.nextOpInfo()[0]) :
         code.addWarning(msgs.STMT_WITH_NO_EFFECT % '--')
+    _modifyStackName(code, '-neg')
 
 def _UNARY_INVERT(oparg, operand, codeSource, code) :
     if OP.UNARY_INVERT(code.nextOpInfo()[0]) :
         code.addWarning(msgs.STMT_WITH_NO_EFFECT % '~~')
+    _modifyStackName(code, '-invert')
 
 
 def _popStackRef(code, operand, count = 2) :
     code.popStackItems(count)
     code.pushStack(Stack.Item(operand, Stack.TYPE_UNKNOWN))
 
+def _popModifiedStack(code, suffix=' '):
+    code.popStack()
+    _modifyStackName(code, suffix)
+
 def _pop(oparg, operand, codeSource, code) :
     code.popStack()
-_POP_TOP = _BINARY_LSHIFT = _BINARY_RSHIFT = \
-           _BINARY_AND = _BINARY_XOR = _BINARY_OR = \
-           _PRINT_ITEM = _pop
+_POP_TOP = _PRINT_ITEM = _pop
+
+def _popModified(oparg, operand, codeSource, code):
+    _popModifiedStack(code)
+_BINARY_LSHIFT = _BINARY_RSHIFT = _popModified
+
+def _checkModifyNoOp(code, op, msg=msgs.MODIFY_VAR_NOOP):
+    stack = code.stack
+    if len(stack) >= 2:
+        name = stack[-1].getName()
+        if name == stack[-2].getName():
+            code.addWarning(msg % (name, op, name))
+
+def _BINARY_AND(oparg, operand, codeSource, code):
+    _checkModifyNoOp(code, '&')
+
+def _BINARY_OR(oparg, operand, codeSource, code):
+    _checkModifyNoOp(code, '|')
+
+def _BINARY_XOR(oparg, operand, codeSource, code):
+    _checkModifyNoOp(code, '^', msgs.XOR_VAR_WITH_ITSELF)
 
 def _PRINT_ITEM_TO(oparg, operand, codeSource, code) :
-    _popStackRef(code, operand)
+    code.popStackItems(2)
 
 try:
     ComplexType = types.ComplexType
@@ -1305,9 +1436,7 @@ _NUMERIC_TYPES = (types.IntType, types.FloatType, ComplexType)
 
 # FIXME: This is pathetically weak, need to handle more types
 def _coerce_type(code) :
-    if OP.POP_TOP(code.nextOpInfo()[0]) :
-        code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
-
+    _checkNoEffect(code)
     newItem = Stack.Item('<stack>', Stack.TYPE_UNKNOWN)
     if len(code.stack) >= 2 :
         s1, s2 = code.stack[-2:]
@@ -1338,9 +1467,7 @@ def _BINARY_SUBTRACT(oparg, operand, codeSource, code) :
 _BINARY_POWER = _BINARY_SUBTRACT
 
 def _BINARY_SUBSCR(oparg, operand, codeSource, code) :
-    if OP.POP_TOP(code.nextOpInfo()[0]) :
-        code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
-
+    _checkNoEffect(code)
     if len(code.stack) >= 2 :
         stack = code.stack
         varType = code.typeMap.get(str(stack[-2].data), [])
@@ -1357,14 +1484,19 @@ def _isint(stackItem, code) :
     return types.IntType in stackTypes
 
 def _BINARY_DIVIDE(oparg, operand, codeSource, code) :
-    if OP.POP_TOP(code.nextOpInfo()[0]) :
-        code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
-
+    _checkNoEffect(code)
+    _checkModifyNoOp(code, '/', msgs.DIVIDE_VAR_BY_ITSELF)
     if cfg().intDivide and len(code.stack) >= 2 :
         if _isint(code.stack[-1], code) and _isint(code.stack[-2], code) :
             code.addWarning(msgs.INTEGER_DIVISION % tuple(code.stack[-2:]))
 
-    code.popStack()
+    _popModifiedStack(code, '/')
+
+def _BINARY_TRUE_DIVIDE(oparg, operand, codeSource, code) :
+    _checkNoEffect(code)
+    _checkVariableOperationOnItself(code, operand, msgs.DIVIDE_VAR_BY_ITSELF)
+    _popModifiedStack(code, '/')
+_BINARY_FLOOR_DIVIDE = _BINARY_TRUE_DIVIDE
 
 def _BINARY_MULTIPLY(oparg, operand, codeSource, code) :
     if len(code.stack) >= 2 :
@@ -1375,20 +1507,24 @@ def _BINARY_MULTIPLY(oparg, operand, codeSource, code) :
         else:
             _coerce_type(code)
     else:
-        code.popStack()
+        _popModifiedStack(code, '*')
 
 def _BINARY_MODULO(oparg, operand, codeSource, code) :
-    if OP.POP_TOP(code.nextOpInfo()[0]) :
-        code.addWarning(msgs.POSSIBLE_STMT_WITH_NO_EFFECT)
-
+    _checkNoEffect(code)
+    if cfg().modulo1 and code.stack and code.stack[-1].data == 1:
+        if len(code.stack) < 2 or \
+           code.stack[-2].getType(code.typeMap) != types.FloatType:
+            code.addWarning(msgs.MODULO_1)
     _getFormatWarnings(code, codeSource)
-    code.popStack()
+    _popModifiedStack(code, '%')
     if code.stack:
         code.stack[-1].const = 0
 
 def _ROT_TWO(oparg, operand, codeSource, code) :
     if len(code.stack) >= 2 :
-        del code.stack[-2]
+        tmp = code.stack[-2]
+        code.stack[-2] = code.stack[-1]
+        code.stack[-1] = tmp
 
 def _SETUP_EXCEPT(oparg, operand, codeSource, code) :
     code.has_except = 1
@@ -1460,7 +1596,7 @@ def _FOR_ITER(oparg, operand, codeSource, code) :
 def _jump(oparg, operand, codeSource, code) :
     if len(code.stack) > 0 :
         topOfStack = code.stack[-1]
-        if topOfStack.isMethodCall(codeSource.classObject, cfg().methodArgName) :
+        if topOfStack.isMethodCall(codeSource.classObject, cfg().methodArgName):
             name = topOfStack.data[-1]
             if codeSource.classObject.methods.has_key(name) :
                 code.addWarning(msgs.USING_METHOD_AS_ATTR % name)
@@ -1501,7 +1637,7 @@ def _is_unreachable(code, topOfStack, branch, if_false) :
 
     # check if we break out of the loop
     i = code.index
-    lastLineNum = code.lastLineNum
+    lastLineNum = code.getLineNum()
     while i < branch :
         op, oparg, i, extended_arg = OP.getInfo(code.bytes, i, extended_arg)
         if OP.LINE_NUM(op) :
@@ -1594,6 +1730,9 @@ DISPATCH[ 22] = _BINARY_MODULO
 DISPATCH[ 23] = _BINARY_ADD
 DISPATCH[ 24] = _BINARY_SUBTRACT
 DISPATCH[ 25] = _BINARY_SUBSCR
+DISPATCH[ 26] = _BINARY_FLOOR_DIVIDE
+DISPATCH[ 27] = _BINARY_TRUE_DIVIDE
+# FIXME: add INPLACE FLOOR/TRUE DIVIDE: 28/29
 DISPATCH[ 31] = _SLICE1
 DISPATCH[ 32] = _SLICE2
 DISPATCH[ 33] = _SLICE3
