@@ -16,6 +16,7 @@ import imp
 from pychecker import OP
 from pychecker import Stack
 from pychecker import function
+from pychecker import python
 
 from pychecker import msgs
 from pychecker import utils
@@ -30,12 +31,23 @@ def _checkSelfArg(method, warnings) :
     """Return a Warning if there is no self parameter or
        the first parameter to a method is not self."""
 
+    if not cfg().methodArgName:
+        return
+
     code = method.function.func_code
     err = None
-    if code.co_argcount < 1 :
+    if method.isStaticMethod():
+        if code.co_argcount > 0 and cfg().methodArgName == code.co_varnames[0]:
+            err = msgs.SELF_IS_ARG % 'staticmethod'
+    elif code.co_argcount < 1: 
         err = msgs.NO_METHOD_ARGS % cfg().methodArgName
-    elif code.co_varnames[0] != cfg().methodArgName :
-        err = msgs.SELF_NOT_FIRST_ARG % cfg().methodArgName
+    else:
+        if method.isClassMethod():
+            if code.co_varnames[0] not in cfg().classmethodArgNames:
+                err = msgs.SELF_NOT_FIRST_ARG % \
+                      (cfg().classmethodArgNames, 'class')
+        elif code.co_varnames[0] != cfg().methodArgName:
+            err = msgs.SELF_NOT_FIRST_ARG % (cfg().methodArgName, '')
 
     if err is not None :
         warnings.append(Warning(code, code, err))
@@ -45,8 +57,8 @@ def _checkNoSelfArg(func, warnings) :
     "Return a Warning if there is a self parameter to a function."
 
     code = func.function.func_code
-    if code.co_argcount > 0 and cfg().methodArgName in code.co_varnames :
-        warnings.append(Warning(code, code, msgs.SELF_IS_ARG))
+    if code.co_argcount > 0 and cfg().methodArgName in code.co_varnames:
+        warnings.append(Warning(code, code, msgs.SELF_IS_ARG % 'function'))
 
 
 def _checkSubclass(c1, c2):
@@ -195,6 +207,9 @@ def _checkFunction(module, func, c = None, main = 0, in_class = 0) :
 
     code = CodeChecks.Code()
     code.init(func)
+    if main:
+        for key in func.function.func_globals.keys():
+            code.unusedLocals[key] = -1
     codeSource = CodeChecks.CodeSource(module, func, c, main, in_class, code)
     try :
         _checkCode(code, codeSource)
@@ -237,7 +252,7 @@ def _checkFunction(module, func, c = None, main = 0, in_class = 0) :
     #   branches in byte code to setup a loop, so subtract off 2/3's of them
     #    / 2 to approximate real branches
     branches = (len(code.branches.keys()) - (2 * code.loops)) / 2
-    lines = (code.lastLineNum - code.func_code.co_firstlineno)
+    lines = (code.getLineNum() - code.func_code.co_firstlineno)
     returns = len(code.returnValues)
     if not main and not in_class :
         args = code.func_code.co_argcount
@@ -378,9 +393,14 @@ def getStandardLibrary() :
         except ImportError :
             return None
 
+def normalize_path(path):
+    return os.path.normpath(os.path.normcase(path))
+
 def removeWarnings(warnings, blacklist, std_lib) :
+    if std_lib is not None:
+        std_lib = normalize_path(std_lib)
     for index in range(len(warnings)-1, -1, -1) :
-        filename = warnings[index].file
+        filename = normalize_path(warnings[index].file)
         if filename in blacklist or (std_lib is not None and
                                      utils.startswith(filename, std_lib)) :
             del warnings[index]
@@ -456,7 +476,12 @@ except NameError:
 
 def _findClassWarnings(module, c, class_code,
                        globalRefs, warnings, suppressions) :
-    classSuppress = getSuppression(str(c.classObject), suppressions, warnings)
+    try:
+        className = str(c.classObject)
+    except TypeError:
+        # goofy __getattr__
+        return
+    classSuppress = getSuppression(className, suppressions, warnings)
     baseClasses = c.allBaseClasses()
     for base in baseClasses :
         baseModule = str(base)
@@ -479,9 +504,31 @@ def _findClassWarnings(module, c, class_code,
         func_code = method.function.func_code
         utils.debug("method:", func_code)
 
-        name = str(c.classObject) + '.' + method.function.func_name
+        try:
+            name = str(c.classObject) + '.' + method.function.func_name
+        except AttributeError:
+            # func_name may not exist
+            continue
         methodSuppress = getSuppression(name, suppressions, warnings)
 
+        if cfg().checkSpecialMethods:
+            funcname = method.function.func_name
+            if funcname[:2] == '__' == funcname[-2:] and \
+               funcname != '__init__':
+                err = None
+                argCount = python.SPECIAL_METHODS.get(funcname, -1)
+                if argCount != -1:
+                    # if the args are None, it can be any # of args
+                    if argCount is not None:
+                        minArgs = maxArgs = argCount
+                        err = CodeChecks.getFunctionArgErr(funcname,
+                                     func_code.co_argcount, minArgs, maxArgs)
+                else:
+                    err = msgs.NOT_SPECIAL_METHOD % funcname
+
+                if err is not None:
+                    warnings.append(Warning(filename, func_code, err))
+                
         if cfg().checkOverridenMethods :
             _checkOverridenMethods(method.function, baseClasses, warnings)
 
@@ -508,7 +555,11 @@ def _findClassWarnings(module, c, class_code,
         err = msgs.UNUSED_MEMBERS % (string.join(memberList, ', '), c.name)
         warnings.append(Warning(filename, c.getFirstLine(), err))
 
-    newStyleClass = issubclass(c.classObject, object)
+    try:
+        newStyleClass = issubclass(c.classObject, object)
+    except TypeError:
+        # FIXME: perhaps this should warn b/c it may be a class???
+        newStyleClass = 0
 
     slots = c.statics.get('__slots__')
     if slots is not None and cfg().slots:
@@ -531,7 +582,13 @@ def _findClassWarnings(module, c, class_code,
         lineNum = coerceMethod.function.func_code.co_firstlineno
         err = msgs.USING_COERCE_IN_NEW_CLASS % c.name
         warnings.append(Warning(filename, lineNum, err))
-        
+
+    gettroMethod = c.methods.get('__getattribute__')
+    if not newStyleClass and gettroMethod:
+        lineNum = gettroMethod.function.func_code.co_firstlineno
+        err = msgs.USING_GETATTRIBUTE_IN_OLD_CLASS % c.name
+        warnings.append(Warning(filename, lineNum, err))
+
     if cfg().noDocClass and c.classObject.__doc__ == None :
         method = c.methods.get(utils.INIT, None)
         if method != None :
@@ -577,8 +634,8 @@ def find(moduleList, initialCfg, suppressions = None) :
         _findFunctionWarnings(module, globalRefs, warnings, suppressions)
 
         for c in module.classes.values() :
-            _findClassWarnings(module, c, classCodes.get(c.name),
-                               globalRefs, warnings, suppressions)
+                _findClassWarnings(module, c, classCodes.get(c.name),
+                                   globalRefs, warnings, suppressions)
 
         if cfg().noDocModule and \
            module.module != None and module.module.__doc__ == None :
@@ -610,3 +667,8 @@ def find(moduleList, initialCfg, suppressions = None) :
     if cfg().ignoreStandardLibrary :
         std_lib = getStandardLibrary()
     return removeWarnings(warnings, getBlackList(cfg().blacklist), std_lib)
+
+if 0:
+    # if you want to test w/psyco, include this
+    import psyco
+    psyco.bind(_checkCode)
